from collections import defaultdict
from typing import (
    Optional,
    Type,
)

from django.core.exceptions import ValidationError
from django.http import HttpResponseRedirect
from django.template import (
    Context,
    Template,
)
from django.utils.translation import gettext
from tri_declarative import (
    EMPTY,
    getattr_path,
    Namespace,
    Refinable,
    setdefaults_path,
)
from tri_struct import Struct

from iommi import (
    Action,
    Column,
    Field,
    Form,
    Fragment,
    MISSING,
    Table,
)
from iommi.base import (
    items,
    values,
)
from iommi.endpoint import path_join
from iommi.table import (
    Cell,
    Cells,
)


class EditCell(Cell):
    def get_path(self):
        return path_join(self.column.iommi_path, str(self.row.pk))

    def render_cell_contents(self):
        field = self.table.edit_form.fields.get(self.column.iommi_name(), None)
        if field:
            path = self.get_path()

            field.initial = MISSING
            field.raw_data = None
            field.value = None
            field.parsed_data = None
            field._errors = set()
            field.form.instance = self.row
            field._iommi_path_override = path
            del field.input.attrs['value']
            field.bind_from_instance()

            input_html = field.input.__html__()

            if self.table.edit_errors:
                errors = self.table.edit_errors.get(path)
                if errors:
                    return Template('{{ input_html }}<br><span class="text-danger"><ul class="errors">{% for error in errors %}<li>{{ error }}</li>{% endfor %}</ul></a>').render(context=Context(dict(input_html=input_html, errors=errors)))

            return input_html
        else:
            return super().render_cell_contents()


class EditCells(Cells):
    class Meta:
        cell_class = EditCell

    def iter_editable_cells(self):
        for column in values(self.iommi_parent().columns):
            if not column.render_column:
                continue

            field = self.iommi_parent().edit_form.fields.get(column.iommi_name(), None)
            if not field:
                continue
            yield self.cell_class(cells=self, column=column)


class EditColumn(Column):
    edit: Field = Refinable()

    def on_refine_done(self):
        super(EditColumn, self).on_refine_done()
        self.edit = None


def edit_table__post_handler(table, request, **_):
    # 1. Validate all the fields
    errors = defaultdict(set)
    parsed_data = {}
    for cells in table.cells_for_rows():
        instance = cells.row
        table.edit_form.instance = instance
        for cell in cells.iter_editable_cells():
            path = cell.get_path()
            field = table.edit_form.fields[cell.column.iommi_name()]
            try:
                parsed_data[path] = field.parse(
                    string_value=request.POST.get(path),
                    **field.iommi_evaluate_parameters(),
                )
            except ValidationError as e:
                errors[path] |= set(e.messages)
            except ValueError as e:
                errors[path] = {str(e)}
            except TypeError as e:
                errors[path] = {str(e)}

    if errors:
        table.edit_errors = errors
        return None

    for cells in table.cells_for_rows():
        instance = cells.row
        table.edit_form.instance = instance
        attrs_to_save = []
        for cell in cells.iter_editable_cells():
            path = cell.get_path()
            value = parsed_data[path]
            field = table.edit_form.fields[cell.column.iommi_name()]
            if getattr_path(instance, field.attr) != value:
                field.write_to_instance(field=field, instance=instance, value=value)
                attrs_to_save.append(field.attr)
        instance.save(update_fields=attrs_to_save)

    if 'post_save' in table.extra:
        table.extra.post_save(**table.iommi_evaluate_parameters())

    return HttpResponseRedirect(request.META['HTTP_REFERER'])


class EditTable(Table):
    edit_errors = None
    edit_form: Form = Refinable()
    form_class: Type[Form] = Refinable()
    parent_form: Optional[Form] = Refinable()

    class Meta:
        form_class = Form
        member_class = EditColumn
        outer__tag = 'form'
        outer__attrs__enctype = 'multipart/form-data'
        outer__attrs__method = 'post'
        cells_class = EditCells
        actions__submit = dict(
            call_target__attribute='primary',
            display_name=gettext('Save'),
            post_handler=edit_table__post_handler,
        )
        actions__csrf = Action(tag='div', children__csrf=Fragment(template=Template('{% csrf_token %}')), attrs__style__display='none')
        actions_below = True
        edit_form = EMPTY

    def on_refine_done(self):
        super(EditTable, self).on_refine_done()

        fields = Struct()

        for name, column in items(self.iommi_namespace.columns):
            if not isinstance(column, EditColumn):
                continue
            if getattr(column, 'include', None) is False:
                continue

            edit_conf = column.iommi_namespace.get('edit', None)

            if not edit_conf:
                continue
            if getattr(column.edit, 'include', None) is False:
                continue

            if isinstance(edit_conf, dict):
                field = setdefaults_path(
                    Namespace(),
                    edit_conf,
                    model=self.model,
                    model_field_name=column.model_field_name,
                    attr=name if column.attr is MISSING else column.attr,
                )
            else:
                field = column.iommi_namespace.edit

            fields[name] = field

        auto = Namespace(self.auto)
        if auto:
            auto.default_included = False

        self.edit_form = self.get_meta().form_class(**setdefaults_path(
            Namespace(),
            self.edit_form,
            fields=fields,
            _name='edit_form',
            auto=auto,
        ))

        declared_fields = self.edit_form.iommi_namespace.fields
        self.edit_form = self.edit_form.refine_defaults(fields=declared_fields).refine_done()

    def on_bind(self) -> None:
        super(EditTable, self).on_bind()
        self.edit_form = self.edit_form.bind(parent=self)
        self._bound_members.edit_form = self.edit_form

        # If this is a nested form register it with the parent, need
        # to do this early because is_target needs self.parent_form
        if self.iommi_parent() is not None:
            self.parent_form = self.iommi_parent().iommi_evaluate_parameters().get('form', None)
            if self.parent_form is not None:
                self.parent_form.nested_forms[self._name] = self
                self.outer.tag = None

    def is_valid(self):
        return self.edit_form.is_valid()

    @property
    def render_actions(self):
        # For now we do not support actions in child forms. This mirrors the behavior in Form
        if self.parent_form:
            return ''

        return super().render_actions



from copy import copy

from tri_declarative import (
    declarative,
    dispatch,
    flatten,
    getattr_path,
    Namespace,
    Refinable,
    refinable,
)

from iommi.base import items


def prefixes(path):
    parts = [p for p in path.split('__') if p]
    for i in range(len(parts)):
        yield '__'.join(parts[: i + 1])


class RefinedNamespace(Namespace):
    __iommi_refined_description: str
    __iommi_refined_parent: Namespace
    __iommi_refined_params: Namespace
    __iommi_refined_defaults: bool

    def __init__(self, description, parent, defaults=False, *args, **kwargs):
        params = Namespace(*args, **kwargs)
        object.__setattr__(self, '__iommi_refined_description', description)
        object.__setattr__(self, '__iommi_refined_parent', parent)
        object.__setattr__(self, '__iommi_refined_params', params)
        object.__setattr__(self, '__iommi_refined_defaults', defaults)
        missing = object()
        if defaults:
            default_updates = Namespace()
            updates = Namespace()
            for path, value in items(flatten(params)):
                found = False
                for prefix in prefixes(path):
                    existing = getattr_path(parent, prefix, missing)
                    if existing is missing:
                        break
                    if isinstance(existing, RefinableObject):
                        new_defaults = getattr_path(params, prefix)
                        if isinstance(new_defaults, dict):
                            existing = existing.refine_defaults(**new_defaults)
                            updates.setitem_path(prefix, existing)
                        else:
                            # We got an entire object here, so throw the rest away!
                            pass
                        found = True
                if not found:
                    default_updates.setitem_path(path, value)
            super().__init__(default_updates, parent, updates)
        else:
            updates = Namespace()
            for path, value in items(flatten(params)):
                found = False
                for prefix in prefixes(path):
                    existing = getattr_path(parent, prefix, missing)
                    if existing is missing:
                        break
                    if isinstance(existing, RefinableObject):
                        new_updates = getattr_path(params, prefix)
                        if isinstance(new_updates, dict):
                            existing = existing.refine(**new_updates)
                        else:
                            existing = new_updates
                        updates.setitem_path(prefix, existing)
                        found = True
                if not found:
                    updates.setitem_path(path, value)
            super().__init__(parent, updates)

    def as_stack(self):
        refinements = []
        default_refinements = []
        node = self

        while isinstance(node, RefinedNamespace):
            try:
                description = object.__getattribute__(node, '__iommi_refined_description')
                parent = object.__getattribute__(node, '__iommi_refined_parent')
                delta = object.__getattribute__(node, '__iommi_refined_params')
                defaults = object.__getattribute__(node, '__iommi_refined_defaults')
                value = (description, flatten(delta))
                if defaults:
                    default_refinements = default_refinements + [value]
                else:
                    refinements = [value] + refinements
                node = parent
            except AttributeError:
                break

        return default_refinements + [('base', flatten(node))] + refinements


def is_refinable_function(attr):
    return getattr(attr, 'refinable', False)


class EvaluatedRefinable(Refinable):
    pass


def evaluated_refinable(f):
    f = refinable(f)
    f.__iommi__evaluated = True
    return f


class RefinableMembers(Refinable):
    pass


def is_evaluated_refinable(x):
    return isinstance(x, EvaluatedRefinable) or getattr(x, '__iommi__evaluated', False)


@declarative(
    member_class=Refinable,
    parameter='refinable',
    is_member=is_refinable_function,
    add_init_kwargs=False,
)
class RefinableObject:

    iommi_namespace: Namespace
    is_refine_done: bool

    @dispatch()
    def __init__(self, namespace=None, **kwargs):
        if namespace is None:
            namespace = Namespace()
        else:
            namespace = Namespace(namespace)

        declared_items = self.get_declared('refinable')
        for name in list(kwargs):
            prefix, _, _ = name.partition('__')
            if prefix in declared_items:
                namespace.setitem_path(name, kwargs.pop(name))

        if kwargs:
            available_keys = '\n    '.join(sorted(declared_items.keys()))
            raise TypeError(
                f"""\
{self.__class__.__name__} object has no refinable attribute(s): {', '.join(f'"{k}"' for k in sorted(kwargs.keys()))}.
Available attributes:
    {available_keys}
"""
            )

        self.iommi_namespace = namespace
        self.is_refine_done = False

    def refine_done(self, parent=None):
        result = copy(self)
        del self

        assert not result.is_refine_done, f"refine_done() already invoked on {result!r}"

        if hasattr(result, 'apply_styles'):
            if parent is None:
                result = result.apply_styles(None, is_root=True)
            else:
                result = result.apply_styles(parent.iommi_style, is_root=False)

        # Apply config from result.namespace to result
        declared_items = result.get_declared('refinable')
        remaining_namespace = Namespace(result.iommi_namespace)
        for k, v in items(declared_items):
            if k == 'iommi_style':
                remaining_namespace.pop(k, None)
                continue
            if isinstance(v, Refinable):
                setattr(result, k, remaining_namespace.pop(k, None))
            else:
                if k in remaining_namespace:
                    setattr(result, k, remaining_namespace.pop(k))

        if remaining_namespace:
            available_keys = '\n    '.join(sorted(declared_items.keys()))
            raise TypeError(
                f"""\
{result.__class__.__name__} object has no refinable attribute(s): {', '.join(f'"{k}"' for k in sorted(remaining_namespace.keys()))}.
Available attributes:
    {available_keys}
"""
            )
        result.is_refine_done = True

        result.on_refine_done()

        return result

    def on_refine_done(self):
        pass

    def refine(self, **args):
        assert not self.is_refine_done, f"Already called refine_done on {self!r}"
        result = copy(self)
        result.iommi_namespace = RefinedNamespace('refine', self.iommi_namespace, **args)
        return result

    def refine_defaults(self, **args):
        assert not self.is_refine_done, f"Already called refine_done on {self!r}"
        result = copy(self)
        result.iommi_namespace = RefinedNamespace('refine defaults', self.iommi_namespace, defaults=True, **args)
        return result

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.iommi_namespace}>"
